# -*- FilePy: FileConfigConvMol -*-
#
# @ Now @ 2017-11-27 18:17:21.158684
#

from f311.filetypes import ConfigDict

ccm = ConfigDict({
 'sourcename': 'Kurucz',
 'fn_output': '',
 'kurucz_iso': None,
 'kurucz_fn_input': '',
 'flag_hlf': False,
 'flag_normhlf': False,
 'flag_fcf': False,
 'flag_special_fcf': False,
 'flag_spinl': False,
 'flag_quiet': False
})
